import { LoadOp, StoreOp } from '../enum.js'
import { CommandRecorderError } from '../utils/errors.js'
import { SwapChain } from './swap-chain.js'

export class CommandRecorder {
  static _defaultClearValue = { r: 0, g: 0, b: 0, a: 1 }

  _device
  _swapChain
  _label

  _encoder

  /** @type {GPURenderPassEncoder | undefined} */
  _passEncoder

  /**
   * @param {GPUDevice} device
   * @param {SwapChain} swapChain
   * @param {string} [label]
   */
  constructor(device, swapChain, label) {
    this._device = device
    this._swapChain = swapChain
    this._label = label

    this._encoder = device.createCommandEncoder({ label })
  }

  /**
   * @returns {[GPURenderPassColorAttachment]}
   */
  _defaultColorAttachment() {
    const view = this._swapChain.getCurrentTextureView()

    return [{
      view,
      clearValue: CommandRecorder._defaultClearValue,
      loadOp: LoadOp.Clear,
      storeOp: StoreOp.Store
    }]
  }
  /**
   * @param {GPURenderPassColorAttachment[]} [colorAttachments]
   * @param {GPURenderPassDepthStencilAttachment} [depthStencilAttachment]
   * @returns {GPURenderPassEncoder}
   */
  beginRenderPass(colorAttachments, depthStencilAttachment) {
    if (this._passEncoder) {
      throw CommandRecorderError.activeRenderPass()
    }

    const attachments = colorAttachments || this._defaultColorAttachment()

    const descriptor = {
      label: this._label || 'RenderPass',
      colorAttachments: attachments,
      depthStencilAttachment
    }

    this._passEncoder = this._encoder.beginRenderPass(descriptor)

    return this._passEncoder
  }

  endRenderPass() {
    if (!this._passEncoder) { return }

    this._passEncoder.end()
    this._passEncoder = undefined
  }

  finish() {
    if (this._passEncoder) {
      this.endRenderPass()
    }

    return this._encoder.finish()
  }
}

