import { SwapChain } from './swap-chain.js'
import { Buffer, UniformBuffer } from '../resources/buffer.js'
import { GraphicsDeviceError, WebGPUError, BufferError, WebGPUObjectError } from '../utils/errors.js'
import { GraphicsDeviceInitialized, GraphicsDeviceLost } from '../utils/events.js'
import { EventEmitter } from '../utils.js'
import { ShaderModule } from '../resources/shader-module.js'
import { RenderPipeline } from '../rendering/render-pipeline.js'
import { CommandRecorder } from './command-recorder.js'
import { BindGroupLayout } from '../resources/bind-group-layout.js'
import { BindGroup } from '../resources/bind-group.js'
import { Texture } from '../resources/texture.js'

/**
 * @typedef BaseBindGroupEntry
 * @property {GPUIndex32} binding
 * @property {BindingResource} resource
 */

/**
 * @typedef {{
     offset?: number
     size?: number
 * }} BufferBindingResource
 *
 * @typedef {{ textureView?: GPUTextureView }} TextureBindingResource
 *
 * @typedef {Buffer | Texture | Sampler} BindingResource
 *
 * @typedef {
     (BufferBindingResource | TextureBindingResource) &
     BaseBindGroupEntry
 * } BindGroupEntry
 */

import { Sampler } from '../resources/sampler.js'
import { ResourceType } from '../utils/internal-enums.js'
import { Material } from '../resources/material.js'

class GraphicsDeviceBuilder {
  _canvas

  get canvas() {
    return this._canvas
  }

  /** @type {GPURequestAdapterOptions} */
  _adapter_options

  /** @type {GPUDeviceDescriptor} */
  _device_descriptor

  /**
   * @param {HTMLCanvasElement} [canvasElement]
   */
  constructor(canvasElement) {
    this._canvas = canvasElement
  }

  isSupported() {
    return navigator.gpu
  }

  /**
   * @param {HTMLCanvasElement} canvasElement
   */
  withCanvas(canvasElement) {
    this._canvas = canvasElement
    return this
  }

  /**
   * @param {GPURequestAdapterOptions} [options]
   */
  withAdapter(options) {
    if (!this.isSupported()) {
      throw WebGPUError.unsupported()
    }

    this._adapter_options = options
    return this
  }

  /**
   * @param {GPUDeviceDescriptor} [options] 
   */
  withDevice(options) {
    if (!this.isSupported()) {
      throw WebGPUError.unsupported()
    }

    this._device_descriptor = options
    return this
  }

  async build() {
    return new GraphicsDevice(
      this._canvas,
      new DeviceHandler(
        this._adapter_options,
        this._device_descriptor
      )
    )
  }
}

class DeviceHandler {
  /** @type {GPURequestAdapterOptions} */
  _adapter_options

  /** @type {GPUAdapter} */
  _adapter

  get adapter() {
    return this._adapter
  }

  /** @type {GPUDeviceDescriptor} */
  _device_descriptor

  /** @type {GPUDevice} */
  _device

  get device() {
    return this._device
  }

  /**
   * @param {GPURequestAdapterOptions} adapterOptions 
   * @param {GPUDeviceDescriptor} deviceDescriptor 
   */
  constructor(adapterOptions, deviceDescriptor) {
    this._adapter_options = adapterOptions
    this._device_descriptor = deviceDescriptor
  }

  async create() {
    this._adapter = await navigator.gpu.requestAdapter(this._adapter_options)

    if (!this._adapter) {
      throw WebGPUError.adapterUnavailable()
    }

    this._device = await this._adapter.requestDevice(this._device_descriptor)

    if (!this._device) {
      throw WebGPUError.deviceUnavailable()
    }
  }
}

export class GraphicsDevice extends EventEmitter {
  _canvas
  _deviceHandler

  /** @type {SwapChain} */
  _swapChain

  /** @type {GPUQueue} */
  _queue

  _isInitialized = false

  get isInitialized() {
    return this._isInitialized
  }

  get adapter() {
    return this._deviceHandler.adapter
  }

  get device() {
    return this._deviceHandler.device
  }

  get queue() {
    return this._queue
  }

  get swapChain() {
    return this._swapChain
  }


  /**
   * @param {HTMLCanvasElement} canvas 
   * @param {DeviceHandler} deviceHandler 
   */
  constructor(canvas, deviceHandler) {
    super()

    this._canvas = canvas
    this._deviceHandler = deviceHandler
  }

  /**
   * @param {HTMLCanvasElement} [canvas]
   */
  static build(canvas) {
    return new GraphicsDeviceBuilder(canvas)
  }

  async initialize() {
    await this._deviceHandler.create()

    this._swapChain = new SwapChain(
      this._canvas,
      this._deviceHandler.device
    )

    this._swapChain.configure()

    this._queue = this.device.queue

    this._deviceHandler.device.lost.then(info => {
      this._isInitialized = false

      this.emit(
        GraphicsDeviceLost.EventName,
        new GraphicsDeviceLost(info)
      )
    })

    this._isInitialized = true

    this.emit(
      GraphicsDeviceInitialized.EventName,
      new GraphicsDeviceInitialized(this)
    )

    return true
  }

  /**
   * @typedef {Omit<GPUBufferDescriptor, 'mappedAtCreation'>} BufferDescriptor
   */

  /**
   * Create a GPU buffer
   * @param {BufferDescriptor} descriptor 
   * @param {ArrayBufferView | ArrayBuffer} [data]
   * @returns {Buffer}
   *
   * @throws {GPUBufferError}
   */
  createBuffer(descriptor, data) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    try {
      const buffer = Buffer.create(this.device, descriptor)

      if (data) {
        buffer.write(data)
      }

      return buffer
    } catch (err) {
      throw BufferError.from(err)
    }
  }

  /**
   * @param {number} size
   * @param {ArrayBufferView | ArrayBuffer} [data]
   * @param {string} [label]
   */
  createUniformBuffer(size, data, label) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    const buffer = UniformBuffer.create(this.device, {
      size,
      label
    })

    if (data) {
      buffer.write(data)
    }

    return buffer
  }

  /**
   * Creates a shader module from WGSL code.
   * @param {string} code
   * @param {string} [label]
   * @param {number} [shaderType]
   * @param {string} [entryPoint]
   * @returns {ShaderModule}
   */
  createShaderModule(code, label, shaderType, entryPoint) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    try {
      return ShaderModule.create(this.device, { code, label, shaderType, entryPoint })
    } catch (err) {
      throw WebGPUObjectError.from(err, ShaderModule)
    }
  }

  /**
     * Creates a render pipeline.
     * @param {GPURenderPipelineDescriptor} descriptor - Raw render pipeline descriptor.
     * @returns {RenderPipeline}
     */
  createRenderPipeline(descriptor) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    try {
      const pipeline = this.device.createRenderPipeline(descriptor)
      return new RenderPipeline(pipeline, descriptor.label)
    } catch (err) {
      throw WebGPUObjectError.from(err, RenderPipeline)
    }
  }

  /**
   * @param {import('../resources/material.js').ShaderPairDescriptor} shaders 
   * @param {BindGroupLayout[]} bindGroupLayouts 
   */
  createMaterial(shaders, bindGroupLayouts) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    try {
      return new Material(this.device, shaders, bindGroupLayouts)
    } catch (err) {
      throw WebGPUObjectError.from(err, Material)
    }
  }

  /**
   * Creates a CommandRecorder to begin recording GPU commands.
   * @param {string} [label]
   * @returns {CommandRecorder}
   */
  createCommandRecorder(label) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    return new CommandRecorder(this.device, this._swapChain, label)
  }

  /**
   * @param {GPUBindGroupLayoutEntry[]} entries
   * @param {string} [label]
   */
  createBindGroupLayout(entries, label) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    return BindGroupLayout.create(this.device, {
      label,
      entries
    })
  }

  /**
   * @param {BindGroupEntry} binding
   * @returns {GPUBindingResource}
   */
  _getBindingResource(binding) {
    const resource = binding.resource
    switch (resource.resourceType) {
      case ResourceType.Sampler:
        return resource.handle

      case ResourceType.TextureView:
        return 'textureView' in binding ?
          binding.textureView :
          resource.getView()

      case ResourceType.Buffer:
        const bufSize = resource.size

        const offset = 'offset' in binding ? binding.offset : 0
        const size = 'size' in binding ?
          binding.size :
          bufSize - offset

        if (offset + size > bufSize) {
          throw BufferError.outOfBounds(offset, size)
        }

        return {
          buffer: resource.handle,
          offset,
          size
        }
    }
  }

  /**
   * @param {BindGroupLayout} layout
   * @param {BindGroupEntry[]} bindings
   * @param {string} [label]
   */
  createBindGroup(layout, bindings, label) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }


    const entries = bindings.map(def => ({
      binding: def.binding,
      resource: this._getBindingResource(def)
    }))

    return BindGroup.create(this.device, {
      layout: layout.handle,
      entries,
      label
    })
  }

  /**
   * @param {Array<BindGroupLayout>} layouts 
   * @param {string} [label]
   */
  createPipelineLayout(layouts, label) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    const bindGroupLayouts = layouts.map(layout => layout.handle)

    return this.device.createPipelineLayout({
      label,
      bindGroupLayouts
    })
  }

  /**
   * @param {GPUSamplerDescriptor} [descriptor]
   */
  createSampler(descriptor) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    return Sampler.create(this.device, descriptor)
  }

  /**
   * @param {GPUTextureDescriptor} descriptor 
   */
  createTexture(descriptor) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    try {
      return Texture.create(this.device, descriptor)
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  /**
   * @param {ImageBitmap} bitmap
   * @param {object} [options]
   * @param {string} [options.label]
   * @param {GPUTextureFormat} [options.format='rgba8unorm']
   * @param {GPUTextureUsageFlags} [options.usage=TEXTURE_BINDING | COPY_DST | RENDER_ATTACHMENT]
   * @param {boolean} [options.generateMipmaps=false]
   * @param {boolean} [options.flipY=false]
   */
  createTextureFromBitmap(bitmap, options) {
    if (!this._isInitialized) { throw GraphicsDeviceError.uninitialized() }

    if (!bitmap) { throw new TypeError('Provided bitmap is null.') }

    const usage = (options.usage ?? (GPUTextureUsage.TEXTURE_BINDING | GPUTextureUsage.COPY_DST | GPUTextureUsage.RENDER_ATTACHMENT)) | GPUTextureUsage.TEXTURE_BINDING | GPUTextureUsage.COPY_DST

    const mipLevelCount = 1

    /** @type {GPUTextureDescriptor} */
    const descriptor = {
      label: options.label,
      size: {
        width: bitmap.width,
        height: bitmap.height,
        depthOrArrayLayers: 1
      },
      format: options.format ?? 'rgba8unorm',
      usage,
      dimension: '2d',
      mipLevelCount,
      sampleCount: 1,
    }

    try {
      const texture = this.device.createTexture(descriptor)
      this.queue.copyExternalImageToTexture(
        { source: bitmap, flipY: options.flipY ?? false },
        { texture, mipLevel: 0, origin: { x: 0, y: 0, z: 0 } },
        descriptor.size
      )
      const wrapper = new Texture(this.device, texture)
      return wrapper
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  /**
   * Submits an array of command buffers to the GPU queue.
   * @param {GPUCommandBuffer[]} commandBuffers
   */
  submitCommands(commandBuffers) {
    if (!this._isInitialized || !commandBuffers || commandBuffers.length === 0) return

    this.queue.submit(commandBuffers)
  }

  /**
   * Cleans up GPU resources. Call when the application exits.
   */
  destroy() {
    if (this.device) {
      this.device.destroy()
    }

    this._deviceHandler = null
    this._queue = null
    this._swapChain = null
    this._isInitialized = false
  }
}

