import { WebGPUError } from '../utils/errors.js'


/** @import { PositiveInteger } from '../utils.js' */

/**
 * @typedef {Omit<GPUCanvasConfiguration, 'device' | 'format'>} SwapChainConfiguration
 */
export class SwapChain {
  _canvas
  _device
  _context
  _format
  _width
  _height

  /** @type {SwapChainConfiguration} */
  _configuration

  get context() {
    return this._context
  }

  get format() {
    return this._format
  }

  get width() {
    return this._width
  }

  get height() {
    return this._height
  }

  /**
   * @param {HTMLCanvasElement} canvas
   * @param {GPUDevice} device
   * @param {GPUCanvasContext} [context]
   * 
   * @throws {WebGPUError}
   * Throws an error if unable to request a WebGPU context
   */
  constructor(canvas, device, context) {
    this._canvas = canvas
    this._device = device

    this._context = context || canvas.getContext('webgpu')

    if (!this._context) {
      throw WebGPUError.contextFailure()
    }

    this._format = navigator.gpu.getPreferredCanvasFormat()
    this._width = canvas.width
    this._height = canvas.height
  }

  /**
   * @param {SwapChainConfiguration} [configuration]
   */
  configure(configuration) {
    if (configuration) {
      this._configuration = configuration
    }

    this._width = this._canvas.width
    this._height = this._canvas.height

    this._context.configure({
      device: this._device,
      format: this._format,
      alphaMode: 'opaque',
      ...this._configuration
    })
  }

  getCurrentTextureView() {
    return this._context.getCurrentTexture().createView()
  }

  /**
   * @template {number} const T
   * @param {PositiveInteger<T>} width
   * @param {PositiveInteger<T>} height
   */
  resize(width, height) {
    if (width <= 0 || height <= 0) {
      return
    }

    if (this._width !== width || this._height !== height) {
      this._canvas.width = width
      this._canvas.height = height
      this.configure()
    }
  }
}

