import { BufferError, WebGPUError } from '../utils/errors.js'
import { ResourceType } from '../utils/internal-enums.js'

/** @import { TypedArray, TypedArrayConstructor } from '../utils.js' */
/** @import { BindGroupEntry, BindingResource, BufferBindingResource } from '../core/graphics-device.js' */

export class Buffer {
  _device
  _handle

  _mapped = false

  /** @type {GPUBuffer} */
  _defaultStagingBuffer

  get handle() {
    return this._handle
  }

  get size() {
    return this._handle.size
  }

  get usage() {
    return this._handle.usage
  }

  get resourceType() {
    return ResourceType.Buffer
  }

  /**
   * @param {GPUDevice} device
   * @param {GPUBuffer} texture
   */
  constructor(device, texture) {
    this._device = device
    this._handle = texture
  }

  /**
   * @param {GPUDevice} device
   * @param {GPUBufferDescriptor} descriptor
   */
  static create(device, descriptor) {
    try {
      return new Buffer(
        device,
        device.createBuffer(descriptor)
      )
    } catch (err) {
      throw BufferError.from(err)
    }
  }

  /**
   * @param {number} [size]
   */
  _createStagingOptions(size = this.size) {
    return {
      size,
      usage: GPUBufferUsage.MAP_READ | GPUBufferUsage.COPY_DST
    }
  }

  /**
   * @param {number} [size]
   */
  _getStagingBuffer(size) {
    return this._device.createBuffer(this._createStagingOptions(size))
  }

  /**
   * @param {ArrayBufferView | ArrayBuffer} data
   * @param {number} [offset=0]
   * @param {number} [dataOffset=0]
   */
  write(data, offset = 0, dataOffset = 0) {
    if (!(this.usage & GPUBufferUsage.COPY_DST)) {
      console.warn('Buffer usage does not include COPY_DST. Buffer.write may fail.')
    }

    if (!this._device?.queue) {
      throw WebGPUError.deviceUnavailable()
    }

    this._device.queue.writeBuffer(
      this._handle,
      offset,
      data,
      dataOffset
    )
  }

  /**
   * @typedef {Exclude<TypedArray, BigInt64Array | BigUint64Array>} SmallTypedArray
   * @typedef {Exclude<TypedArrayConstructor, BigInt64ArrayConstructor | BigUint64ArrayConstructor>} SmallTypedArrayConstructor
   */

  /**
   * @param {SmallTypedArray | DataView | undefined} [out]
   * @param {number} [byteOffset=0]
   * @param {number} [byteSize]
   */
  async read(out, byteOffset = 0, byteSize = -1) {
    if (!this._device) {
      throw WebGPUError.deviceUnavailable()
    }

    if (!(this.usage & GPUBufferUsage.MAP_READ)) {
      throw BufferError.invalidRead()
    }

    if (byteOffset < 0) { throw new RangeError('Read byteOffset cannot be negative') }
    if (byteSize < 0) { byteSize = this.size - byteOffset }
    if (byteSize < 0) { throw new RangeError(`Invalid calculated byteSize (${byteSize})`) }
    if (byteSize === 0) { return out ?? new ArrayBuffer(0) }
    if (byteOffset + byteSize > this.size) { throw new RangeError(`Read range exceeds buffer size`) }


    if (out != null) {
      if (!ArrayBuffer.isView(out)) { throw new TypeError('"out" parameter must be a TypedArray or DataView.') }

      if (out.byteLength < byteSize) { throw new RangeError(`Provided output buffer too small`) }
    }

    let result
    let range

    try {
      await this.handle.mapAsync(GPUMapMode.READ, byteOffset, byteSize)
      range = this.handle.getMappedRange(byteOffset, byteSize)

      if (out != null) {
        const SourceView = /** @type {SmallTypedArrayConstructor} */ (out.constructor)
        const bytesPerElement = SourceView.BYTES_PER_ELEMENT

        if (!bytesPerElement) {
          if (out instanceof DataView) {
            new Uint8Array(
              out.buffer,
              out.byteOffset,
              byteSize
            ).set(new Uint8Array(range))
          } else {
            throw new TypeError('"out" is not a standard TypedArray or DataView')
          }
        } else {
          if (byteSize % bytesPerElement !== 0) {
            throw new RangeError(`"byteSize" (${byteSize}) incompatible with "out" byte size (${bytesPerElement})`)
          }

          const view = new SourceView(range)
          const target = new SourceView(
            out.buffer,
            out.byteOffset,
            byteSize / bytesPerElement
          )

          target.set(view)
        }

        result = out
      } else {
        result = range.slice(0)
      }
    } catch (err) {
      throw BufferError.from(err)
    } finally {
      this.handle.unmap()
    }

    return result
  }

  destroy() {
    this._handle?.destroy()
    this._handle = null
  }
}


export class UniformBuffer extends Buffer {
  /**
   * @param {GPUDevice} device
   * @param {GPUBuffer} buffer
   */
  constructor(device, buffer) {
    super(device, buffer)
  }

  /**
   * @param {GPUDevice} device
   * @param {Omit<GPUBufferDescriptor, 'usage'>} descriptor
   */
  static create(device, descriptor) {
    const usage = GPUBufferUsage.UNIFORM | GPUBufferUsage.COPY_DST
    return super.create(device, {
      usage,
      ...descriptor
    })
  }
}

