import { ResourceType, TemplateInfo, VariableInfo } from 'wgsl_reflect'
import { BindGroupLayout } from './bind-group-layout.js'
import { ShaderModule } from './shader-module.js'
import { MaterialError } from '../utils/errors.js'
import { GroupBindingMap } from '../utils/bindings.js'
import { accessToBufferType, accessToStorageTextureAccess, parseTextureType, typeToViewDimension, wgslToWgpuFormat } from '../utils/wgsl-to-wgpu.js'

/** @import {Either} from '../utils.js' */
/** @import { WGSLAccess, WGSLTextureType } from '../utils/wgsl-to-wgpu.js' */

/**
 * @typedef Shader2
 * @property {ShaderModule} vertex
 * @property {ShaderModule} fragment
 *
 * @typedef {Omit<Shader2, 'fragment'>} VertexOnly
 * @typedef {Omit<Shader2, 'vertex'>} FragmentOnly
 *
 * @typedef Unified1
 * @property {ShaderModule} shaderModule
 */

/**
 * @typedef {
     (Shader2 | Either<VertexOnly, FragmentOnly>) |
     Unified1
 * } ShaderPairDescriptor
 */

export class Material {
  _device
  _shaders
  _bindGroupLayouts
  _pipelineLayout

  get shaders() {
    return this._shaders
  }

  get bindGroupLayouts() {
    return this._bindGroupLayouts
  }

  /**
   * @param {GPUDevice} device
   * @param {ShaderPairDescriptor} shaders
   * @param {BindGroupLayout[]} bindGroupLayouts
   */
  constructor(device, shaders, bindGroupLayouts) {
    this._device = device
    this._shaders = Material._parseShaders(shaders)
    this._bindGroupLayouts = bindGroupLayouts

    if (bindGroupLayouts && bindGroupLayouts.length > 0) {
      this._pipelineLayout = device.createPipelineLayout({
        bindGroupLayouts: bindGroupLayouts.map(bgl => bgl.handle)
      })
    }
  }
  /**
   * Attempts to handle shader modules which represent multiple
   * shader types.
   *
   * @param {ShaderPairDescriptor} shaders
   * @returns {UnifiedShader | ShaderPair}
   */
  static _parseShaders(shaders) {
    if (shaders == null) {
      throw MaterialError.missingShader('both')
    }

    if (shaders instanceof ShaderModule) {
      return new UnifiedShader(shaders)
    }

    const hasVertex = 'vertex' in shaders
    const hasFragment = 'fragment' in shaders

    if (hasVertex && hasFragment) {
      return new ShaderPair(shaders.vertex, shaders.fragment)
    } else if (!hasFragment) {
      throw MaterialError.missingShader('fragment')
    } else if (!hasVertex) {
      throw MaterialError.missingShader('vertex')
    }
  }
}
