import { ResourceType, VariableInfo, WgslReflect } from 'wgsl_reflect'
import { WebGPUObjectError } from '../utils/errors.js'
import { BindingMap, GroupBindingMap, VariableStageInfo } from '../utils/bindings.js'

import {
  accessToBufferType,
  accessToStorageTextureAccess,
  parseTextureType,
  typeToViewDimension,
  wgslToWgpuFormat
} from '../utils/wgsl-to-wgpu.js'

/** @import { WGSLAccess, WGSLSampledType, WGSLSamplerType, WGSLTextureType } from '../utils/wgsl-to-wgpu.js' */

export class ShaderModule {
  _handle
  _code

  /** @type {WgslReflect | undefined} */
  _reflection

  get handle() {
    return this._handle
  }

  get label() {
    return this._handle.label
  }

  /**
   * @param {GPUDevice} device
   * @param {string} code
   * @param {string} [label]
   */
  constructor(device, code, label) {
    this._code = code

    try {
      this._handle = device.createShaderModule({ code, label })
    } catch (err) {
      throw WebGPUObjectError.from(err, ShaderModule)
    }
  }

  reflect() {
    if (this._reflection == null) {
      this._reflection = new WgslReflect(this._code)
      this._code = undefined
    }

    return this._reflection
  }
}

export class ReflectedShader {
  /**
   * @param {WgslReflect} reflection
   * @param {GPUShaderStageFlags} stages
   * @param {GroupBindingMap} [out=new GroupBindingMap()]
   */
  _getBindingsForStage(reflection, stages, out = new GroupBindingMap()) {
    const groups = reflection.getBindGroups()

    groups.forEach((bindings, groupIndex) => {
      if (!out.has(groupIndex)) {
        out.set(groupIndex, new BindingMap())
      }

      const bindingsMap = out.get(groupIndex)

      bindings.filter(x => x != null).forEach(variableInfo => {
        if (!bindingsMap.has(variableInfo.binding)) {
          bindingsMap.set(variableInfo.binding, { variableInfo, stages })
        } else {
          bindingsMap.get(variableInfo.binding).stages |= stages
        }
      })
    })

    return out
  }
  /**
   * @param {Map<any, any>} map 
   * @returns {number[]}
   */
  _sortKeyIndices(map) {
    return Array.from(map.keys()).sort((a, b) => a - b)
  }

  /**
   * @param {VariableInfo} _variableInfo
   * @returns {GPUBufferBindingLayout}
   */
  _parseUniform(_variableInfo) {
    return {
      type: 'uniform',
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  /**
   * @param {VariableInfo} variableInfo
   * @returns {GPUBufferBindingLayout}
   */
  _parseStorage(variableInfo) {
    return {
      type: accessToBufferType(
        /** @type {WGSLAccess} */
        (variableInfo.access)
      ),
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  /**
   * @param {WGSLTextureType} type 
   * @param {WGSLSampledType} sampledType 
   * @returns {GPUTextureSampleType}
   */
  _parseSampleType(type, sampledType) {
    if (type.includes('depth')) {
      return 'depth'
    }

    switch (sampledType) {
      case 'f32':
      case 'i32':
      case 'u32':
      default:
        return 'float'
    }
  }

  /**
   * @param {VariableInfo} variableInfo
   * @returns {GPUTextureBindingLayout}
   */
  _parseTexture(variableInfo) {
    const [type, sampledType] = parseTextureType(variableInfo.type.name)

    return {
      sampleType: this._parseSampleType(type, sampledType),
      viewDimension: typeToViewDimension(type),
      multisampled: type.includes('multisampled')
    }
  }

  /**
   * @param {WGSLSamplerType} type
   * @returns {GPUSamplerBindingType}
   */
  _parseSamplerType(type) {
    switch (type) {
      case 'sampler_comparison':
        return 'comparison'
      case 'sampler':
      default:
        return 'filtering'
    }
  }

  /**
   * @param {VariableInfo} variableInfo 
   * @returns {GPUSamplerBindingLayout}
   */
  _parseSampler(variableInfo) {
    return {
      type: this._parseSamplerType(
        /** @type {WGSLSamplerType} */(variableInfo.type.name)
      )
    }
  }

  /**
   * @param {VariableInfo} variableInfo 
   * @returns {GPUStorageTextureBindingLayout}
   */
  _parseStorageTexture(variableInfo) {
    const [type] = parseTextureType(variableInfo.type.name)

    return {
      access: accessToStorageTextureAccess(
        /** @type {WGSLAccess} */(variableInfo.access)
      ),
      format: wgslToWgpuFormat(variableInfo.type.name),
      viewDimension: typeToViewDimension(type)
    }
  }

  /**
   * @param {VariableStageInfo} variableStageInfo
   * @returns {GPUBindGroupLayoutEntry}
   */
  _variableInfoToEntry(variableStageInfo) {
    const { stages: visibility, variableInfo } = variableStageInfo

    switch (variableInfo.resourceType) {
      case ResourceType.Uniform:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseUniform(variableInfo)
        }
      case ResourceType.Storage:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseStorage(variableInfo)
        }
      case ResourceType.Texture:
        return {
          binding: variableInfo.binding,
          visibility,
          texture: this._parseTexture(variableInfo)
        }
      case ResourceType.Sampler:
        return {
          binding: variableInfo.binding,
          visibility,
          sampler: this._parseSampler(variableInfo)
        }
      case ResourceType.StorageTexture:
        return {
          binding: variableInfo.binding,
          visibility,
          storageTexture: this._parseStorageTexture(variableInfo)
        }
      default:
        console.warn(`Unsupported resource type for reflection: ${ResourceType[variableInfo.resourceType]}`)
        return
    }
  }

  /**
   * @param {GroupBindingMap} groupBindings
   */
  _createBindGroupLayoutEntries(groupBindings) {
    const sortedGroupIndices = this._sortKeyIndices(groupBindings)

    return sortedGroupIndices.map(groupIndex => {
      const bindingsMap = groupBindings.get(groupIndex)
      return this._sortKeyIndices(bindingsMap).map(bindingIndex => (
        this._variableInfoToEntry(bindingsMap.get(bindingIndex))
      ))
    })
      .filter(desc => desc.length > 0)
  }
}

export class UnifiedShader extends ReflectedShader {
  _shader

  /**
   * @param {ShaderModule} shader
   */
  constructor(shader) {
    super()

    this._shader = shader
  }

  createBindGroupLayoutEntries() {
    return this._createBindGroupLayoutEntries(
      this._getBindingsForStage(
        this._shader.reflect(),
        GPUShaderStage.VERTEX | GPUShaderStage.FRAGMENT
      )
    )
  }
}

export class ReflectedShaderPair extends ReflectedShader {
  _vertex
  _fragment

  /**
   * @param {ShaderModule} vertex
   * @param {ShaderModule} [fragment]
   */
  constructor(vertex, fragment) {
    super()

    this._vertex = vertex
    this._fragment = fragment
  }

  _createGroupBindings() {
    const groupBindings = new GroupBindingMap()
    this.getBindingsForStage(
      this._vertex.reflect(),
      GPUShaderStage.VERTEX,
      groupBindings
    )

    this.getBindingsForStage(
      this._fragment.reflect(),
      GPUShaderStage.FRAGMENT,
      groupBindings
    )

    return groupBindings
  }

  createBindGroupLayoutEntries() {
    return this._createBindGroupLayoutEntries(
      this._createGroupBindings()
    )
  }
}

