/**
 * @template T, U
 * @typedef {
     { [P in keyof T]: T[P] } &
     { [P in keyof U]?: never}
 * } Only
 */

/**
 * @template T, U
 * @typedef {Only<T, U> | Only<U, T>} Either
 */

/**
 * @template T
 * @typedef {new (...args: any[]) => T} Newable<T>
 */

/**
 * @typedef {
    Int8Array
  | Uint8Array
  | Uint8ClampedArray
  | Int16Array
  | Uint16Array
  | Int32Array
  | Uint32Array
  | Float32Array
  | Float64Array
  | BigInt64Array
  | BigUint64Array
  } TypedArray
 */

/**
 * @typedef {
    Int8ArrayConstructor
  | Uint8ArrayConstructor
  | Uint8ClampedArrayConstructor
  | Int16ArrayConstructor
  | Uint16ArrayConstructor
  | Int32ArrayConstructor
  | Uint32ArrayConstructor
  | Float32ArrayConstructor
  | Float64ArrayConstructor
  | BigInt64ArrayConstructor
  | BigUint64ArrayConstructor
  } TypedArrayConstructor
 */

/**
 * @typedef {
    [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024,
    2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576,
    2097152, 4194304, 8388608, 16777216, 33554432, 67108864, 134217728, 268435456,
    536870912, 1073741824, 2147483648]
 * } PowersOfTwo
 */

/**
 * @template {number} T
 * @typedef {`${T}` extends `-${string}` | '0' | `${string}.${string}` ? never : T } PositiveInteger
 */

/**
 * @template {ReadonlyArray<string>} T
 * @typedef {T extends readonly [] ? [] : T extends readonly [infer Head extends string, ... infer Tail extends ReadonlyArray<string>] ? [Capitalize<Head>, ...CapitalizeAll<Tail>] : string[]} CapitalizeAll
 */

/**
 * @template {ReadonlyArray<string>} T
 * @template {string} [Sep='']
 * @typedef {T extends readonly [] ? '' : T extends readonly [infer Head] ? Head : T extends readonly [infer Head extends string, ...infer Tail extends ReadonlyArray<string>] ? `${Head}${Sep}${Join<Tail, Sep>}` : string} Join
 */

/**
 * @template {string} S
 * @template {string} D
 * @typedef {D extends '' ? [S] : S extends `${infer Head}${D}${infer Tail}` ? [Head, ...Split<Tail, D>] : [S]} Split
 */

/*
 * @typedef {'-' | '_' | ' '} WordSeparator
 */

/**
 * @template {string} S
 * @template {string} D
 * @typedef {S extends `${infer Left}${D}${infer Right}` ? `${PascalCaseFromDelimiter<Left, D>}${Capitalize<PascalCaseFromDelimiter<Right, D>>}` : S} PascalCaseFromDelimiter
 */

/**
 * @template {string} S
 * @typedef {Capitalize<PascalCaseFromDelimiter<PascalCaseFromDelimiter<PascalCaseFromDelimiter<Lowercase<S>, '-'>, '_'>, ' '>>} PascalCaseString
 */

/**
 * @template {string | ReadonlyArray<string>} T
 * @typedef {T extends String ? PascalCaseString<T> : T extends ReadonlyArray<string> ? Join<CapitalizeAll<T>, ''> : T} PascalCase
 */

/**
 * @template {readonly any[]} T
 * @template V
 * @template {any[]} [Acc=[]]
 * @typedef {
     T extends readonly [infer Head, ...infer Tail]
     ? Head extends V ? Acc['length'] : IndexOf<Tail, V, [...Acc, Head]>
     : never
 * } IndexOf
 */

/**
 * @template {string} T
 * @param {T} s
 * @returns {Uppercase<T>}
 */
export const uppercase = s => /** @type {Uppercase<T>} */(s.toUpperCase())

/**
 * @template {string} T
 * @param {T} s
 * @returns {Lowercase<T>}
 */
export const lowercase = s => /** @type {Lowercase<T>} */(s.toLowerCase())

/**
 * @template {string} const T
 * @param {T} s
 * @returns {Split<T, '-'>}
 */
export const fromKebab = s => /** @type {Split<T, '-'>} */(s.split('-'))

/**
 * @template {string} const T
 * @param {T} xs
 * @returns {PascalCase<T>}
 */
export const toPascal = xs => /** @type {PascalCase<T>} */(uppercase(xs[0]) + xs.slice(1))

/**
 * @template {string[]} const T
 * @param {T} xs
 * @returns {PascalCase<T>}
 */
const pascal = xs => /** @type {PascalCase<T>} */(xs.map(toPascal).join(''))

/**
 * @template {string} const T
 * @param {...T} values
 * @returns {Readonly<{ [K in T as PascalCase<K>]: K }>}
 */
export const Enum = (...values) => /** @type {Readonly<{ [K in T as PascalCase<K>]: K }>} */(Object.freeze(
  values.reduce((acc, x) => {
    const key = pascal(fromKebab(x)).toString()
    acc[key] = x
    return acc
  }, {})
))

/**
 * @template T
 * @typedef {T extends `${infer N extends number}` ? N : never } ParseInt
 */

/**
 * @template {string} const T
 * @template {T[]} const A
 * @param {A} values
 * @returns {Readonly<{ [K in keyof A as A[K] extends T ? PascalCase<A[K]> : never]: PowersOfTwo[ParseInt<K>] }>}
 */
export const FlagEnum = (...values) => /** @type {never} */(Object.freeze(
  values.reduce((acc, x, i) => {
    const key = pascal(fromKebab(x)).toString()
    acc[key] = 1 << i
    return acc
  }, {})
))


/** @typedef {(...args: any) => void} Listener */
export class EventEmitter {
  _listeners = {}

  /**
   * @param {PropertyKey} event 
   * @param {Listener} callback 
   */
  on(event, callback) {
    this._listeners[event] = this._listeners[event] || []
    this._listeners[event].push(callback)
  }

  /**
   * @param {PropertyKey} event 
   * @param {...any} args 
   */
  emit(event, ...args) {
    const listeners = this._listeners[event]

    if (listeners) {
      listeners.forEach(
        /** @param {Listener} cb */
        cb => cb(...args)
      )
    }
  }

  /**
   * @param {PropertyKey} event 
   * @param {Listener} callback 
   */
  off(event, callback) {
    const listeners = this._listeners[event]

    if (listeners) {
      this._listeners[event] = listeners.filter(
        /** @param {Listener} cb */
        cb => cb !== callback
      )
    }
  }
}

/**
 * @template {{}} T
 * @template {keyof T} K
 * @param {T} obj
 * @param {...K} keys
 * @returns {Pick<T, K>}
 */
export const pick = (obj, ...keys) => /** @type {Pick<T, K>} */(Object.fromEntries(
  keys
    .filter(key => key in obj)
    .map(key => [key, obj[key]])
))

/**
 * @template {{}} T
 * @template {PropertyKey & keyof T} K
 * @param {T} obj
 * @param {...K} keys
 * @returns {{ [Key in K]: Key extends keyof T ? T[Key] : never }}
 */
export const inclusivePick = (obj, ...keys) => /** @type {{[Key in K]: Key extends keyof T ? T[Key] : never}} */(Object.fromEntries(
  keys.map(key => [key, obj[key]])
))

/**
 * @template {{}} T
 * @template {PropertyKey & keyof T} K
 * @param {T} obj
 * @param {...K} keys
 * @returns {Omit<T, K>}
 */
export const omit = (obj, ...keys) => /** @type {Omit<T, K>} */(Object.fromEntries(
  Object.entries(obj)
    .filter(([key]) => !keys.includes(/** @type {K} */(key)))
))

/** @template {number} T */
export class Flags {
  _value

  /**
   * @param {T} value
   */
  constructor(value) {
    this._value = value
  }

  /**
   * Tests if `flag` exists in `bitflags`
   *
   * @template {number} T
   * @param {T} flag
   * @param {T} bitflags
   */
  static has(flag, bitflags) {
    return bitflags & flag
  }

  /**
   * Combine two flags
   *
   * @template {number} T
   * @param {T} a
   * @param {T} b
   */
  static add(a, b) {
    return a | b
  }

  /**
   * Tests if `flag` exists in these bitflags
   *
   * @param {T} flag
   */
  has(flag) {
    return Flags.has(flag, this._value)
  }

  /**
   * Add `flag` to these bitflags
   * @param {T} flag
   */
  add(flag) {
    return Flags.add(flag, this._value)
  }
}

