/** @import { Newable } from '../utils.js' */

export class WebGPUError extends Error {
  /**
   * @param {string} message 
   * @param {ErrorOptions} [options] 
   */
  constructor(message, options) {
    super(`WebGPUError: ${message}`, options)
  }

  static unsupported() {
    return new WebGPUError('WebGPU is not supported on this browser')
  }

  static contextFailure() {
    return new WebGPUError('Failed to request WebGPU context')
  }

  static adapterUnavailable() {
    return new WebGPUError('Failed to request GPU adapter')
  }

  static deviceUnavailable() {
    return new WebGPUError('Failed to request GPU device')
  }
}

export class GraphicsDeviceError extends Error {
  static uninitialized() {
    return new GraphicsDeviceError('GraphicsDeviceError: GraphicsDevice not initialized')
  }
}

export class CommandRecorderError extends Error {
  static activeRenderPass() {
    return new CommandRecorderError('CommandRecorderError: can\'t begin new render pass while another is active. call CommandRecorder.end()')
  }
}

export class WebGPUObjectError extends Error {
  /**
   * @template T
   * @param {Error} cause
   * @param {string | Newable<T>} [type]
   */
  static from(cause, type) {
    const name = typeof type === 'string' ? type : type.name
    return new WebGPUObjectError(`could not create ${name}`, { cause })
  }
}

export class BufferError extends WebGPUObjectError {
  /**
   * @param {string} message 
   * @param {ErrorOptions} [options] 
   */
  constructor(message, options) {
    super(`BufferError: ${message}`, options)
  }

  /**
   * @param {Error} cause 
   */
  static from(cause) {
    return new BufferError('could not create buffer', { cause })
  }

  static invalidRead() {
    return new BufferError('cannot read a buffer without MAP_READ usage')
  }

  /**
   * @param {number} offset
   * @param {number} size
   */
  static outOfBounds(offset, size) {
    return new BufferError(`buffer offset/size (${offset}/${size}) exceeds buffer dimensions`)
  }
}

export class MaterialError extends WebGPUObjectError {
  /**
   * @param {string} message 
   * @param {ErrorOptions} [options] 
   */
  constructor(message, options) {
    super(`MaterialError: ${message}`, options)
  }

  /**
   * @param {Error} cause 
   */
  static from(cause) {
    return new BufferError('could not create material', { cause })
  }

  /**
   * @param {string} shaderType
   */
  static missingShader(shaderType) {
    return new BufferError(`missing ${shaderType} shader`)
  }
}
