/**
 * @typedef {'read' | 'write' | 'read_write'} WGSLAccess
 * @typedef {'f32' | 'i32' | 'u32'} WGSLSampledType
 */

import { BufferBindingType, StorageTextureAccess, TextureFormat } from '../enum.js'

/**
 * @typedef {
     'texture_1d'
     | 'texture_2d'
     | 'texture_2d_array'
     | 'texture_3d'
     | 'texture_cube'
     | 'texture_cube_array'
 * } WGSLSampledTextureType
 *
 * @typedef {
     'texture_multisampled_2d'
     | 'texture_depth_multisampled_2d'
 * } WGSLMultisampledTextureType
 *
 * @typedef {
     'texture_storage_1d'
     | 'texture_storage_2d'
     | 'texture_storage_2d_array'
     | 'texture_storage_3d'
 * } WGSLStorageTextureType
 *
 * @typedef {
     'texture_depth_2d'
     | 'texture_depth_2d_array'
     | 'texture_depth_cube'
     | 'texture_depth_cube_array'
 * } WGSLDepthTextureType
 *
 * @typedef {
     WGSLSampledTextureType
     | WGSLMultisampledTextureType
     | WGSLStorageTextureType
     | WGSLDepthTextureType
 * } WGSLTextureType
 */

/**
 * @typedef {
     'sampler'
     | 'sampler_comparison'
 * } WGSLSamplerType
 */

/**
 * @param {string} typeName
 * @returns {[WGSLTextureType, WGSLSampledType]}
 */
export const parseTextureType = (typeName) => {
  const chevronIndex = typeName.indexOf('<')
  const type = typeName.slice(0, chevronIndex)
  const sampledType = typeName.slice(chevronIndex + 1, -1)
  return [
      /** @type {WGSLTextureType} */ (type),
      /** @type {WGSLSampledType} */ (sampledType)
  ]
}

/**
 * @param {WGSLAccess} access
 * @returns {GPUBufferBindingType}
 */
export const accessToBufferType = access => {
  switch (access) {
    case 'read': return BufferBindingType.ReadOnlyStorage
    case 'write':
    case 'read_write':
    default:
      return BufferBindingType.Storage
  }
}

/**
 * @param {WGSLAccess} access
 * @returns {GPUStorageTextureAccess}
 */
export const accessToStorageTextureAccess = access => {
  switch (access) {
    case 'read': return StorageTextureAccess.ReadOnly
    case 'write': return StorageTextureAccess.WriteOnly
    case 'read_write': return StorageTextureAccess.ReadWrite
  }
}

export const FormatToFilterType = Object.freeze({
  r8unorm: 'float',
  r8snorm: 'float',
  r8uint: 'uint',
  r8sint: 'sint',
  r16uint: 'uint',
  r16sint: 'sint',
  r16float: 'float',
  rg8unorm: 'float',
  rg8snorm: 'float',
  rg8uint: 'uint',
  rg8sint: 'sint',
  r32uint: 'uint',
  r32sint: 'sint',
  r32float: 'unfilterable-float',
  rg16uint: 'uint',
  rg16sint: 'sint',
  rg16float: 'float',
  rgba8unorm: 'float',
  'rgba8unorm-srgb': 'float',
  rgba8snorm: 'float',
  rgba8uint: 'uint',
  rgba8sint: 'sint',
  bgra8unorm: 'float',
  'bgra8unorm-srgb': 'float',
  rgb10a2unorm: 'float',
  rg11b10ufloat: 'float',
  rgb9e5ufloat: 'float',
  rg32uint: 'uint',
  rg32sint: 'sint',
  rg32float: 'unfilterable-float',
  rgba16uint: 'uint',
  rgba16sint: 'sint',
  rgba16float: 'float',
  rgba32uint: 'uint',
  rgba32sint: 'sint',
  rgba32float: 'unfilterable-float',

  depth16unorm: 'depth',
  depth24plus: 'depth',
  'depth24plus-stencil8': 'depth',
  depth32float: 'unfilterable-float',
  'depth32float-stencil8': 'unfilterable-float',
  stencil8: 'uint'
})

/** @param {string} format */
export const formatToFilterType = format => (
  FormatToFilterType[format] || 'float'
)

export const FormatToStride = Object.freeze({
  r8unorm: 1,
  r8snorm: 1,
  r8uint: 1,
  r8sint: 1,
  r16uint: 2,
  r16sint: 2,
  r16float: 2,
  rg8unorm: 2,
  rg8snorm: 2,
  rg8uint: 2,
  rg8sint: 2,
  r32uint: 4,
  r32sint: 4,
  r32float: 4,
  rg16uint: 4,
  rg16sint: 4,
  rg16float: 4,
  rgba8unorm: 4,
  'rgba8unorm-srgb': 4,
  rgba8snorm: 4,
  rgba8uint: 4,
  rgba8sint: 4,
  bgra8unorm: 4,
  'bgra8unorm-srgb': 4,
  rgb10a2unorm: 4,
  rg11b10ufloat: 4,
  rgb9e5ufloat: 4,
  rg32uint: 4,
  rg32sint: 4,
  rg32float: 4,
  rgba16uint: 4,
  rgba16sint: 4,
  rgba16float: 4,
  rgba32uint: 4,
  rgba32sint: 4,
  rgba32float: 4,

  depth16unorm: 1,
  depth24plus: 1,
  'depth24plus-stencil8': 1,
  depth32float: 1,
  'depth32float-stencil8': 1,
  stencil8: 1
})

/** @param {string} typeName */
export const typeToStride = typeName => (
  FormatToStride[typeName] || 1
)


/** @param {WGSLTextureType} typeName */
export const typeToViewDimension = typeName => {
  switch (typeName) {
    case 'texture_1d':
    case 'texture_storage_1d':
      return '1d'

    case 'texture_2d':
    case 'texture_storage_2d':
    case 'texture_multisampled_2d':
    case 'texture_depth_2d':
    case 'texture_depth_multisampled_2d':
      return '2d'

    case 'texture_2d_array':
    case 'texture_storage_2d_array':
    case 'texture_depth_2d_array':
      return '2d-array'

    case 'texture_3d':
    case 'texture_storage_3d':
      return '3d'

    case 'texture_cube':
    case 'texture_depth_cube':
      return 'cube'

    case 'texture_cube_array':
    case 'texture_depth_cube_array':
      return 'cube-array'

    default:
      return '2d'
  }
}

/**
 * @param {GPUTextureViewDimension} dimension
 * @returns {GPUTextureDimension}
 */
export const textureToImageDimension = dimension => {
  switch (dimension) {
    case '1d':
      return '1d'

    case '2d':
    case '2d-array':
    case 'cube-array':
      return '2d'

    case '3d':
      return '3d'

    default:
      return '2d'
  }
}

/**
 * @param {string} format
 * @returns {GPUTextureFormat}
 */
export const wgslToWgpuFormat = (format) => {
  switch (format) {
    case 'f32':
      return TextureFormat.Bgra8unorm
    case 'vec4<f32>':
    case 'vec4f':
      return TextureFormat.Bgra8unorm
    case 'vec4<u32>':
    case 'vec4u':
      return TextureFormat.Rgba32uint
    case 'vec4<i32>':
    case 'vec4i':
      return TextureFormat.Rgba32sint
    case 'vec2<f32>':
    case 'vec2f':
      return TextureFormat.Rg32float
    case 'u32':
    case 'u':
      return TextureFormat.R32uint
    case 'i32':
    case 'i':
      return TextureFormat.R32sint
    case 'f16':
      return TextureFormat.Rgba16float
    case 'vec4<f16>':
      return TextureFormat.Rgba16float
    case 'vec2<f16>':
      return TextureFormat.Rg16float
    case 'u16':
      return TextureFormat.R16uint
    case 'i16':
      return TextureFormat.R16sint
    case 'vec4unorm':
      return TextureFormat.Rgba8unorm
    case 'vec4snorm':
      return TextureFormat.Rgba8snorm
    case 'vec4<u8>':
      return TextureFormat.Rgba8uint
    case 'vec4<i8>':
      return TextureFormat.Rgba8sint
    default:
      return TextureFormat.Rgba8unorm
  }
}

